package earth.krakatao;

import earth.krakatao.events.KraSocketClientEventInitiater;
import earth.krakatao.protocol.KraSocketClientProtocol;
import earth.krakatao.protocol.KraSocketClientProtocolDest;
import earth.krakatao.protocol.KraSocketClientProtocolMessage;
import earth.krakatao.protocol.KraSocketClientProtocolStatus;
import java.net.URI;
import java.nio.ByteBuffer;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;
import java.util.UUID;
import lombok.Getter;
import lombok.Setter;
import org.java_websocket.client.WebSocketClient;
import org.java_websocket.exceptions.WebsocketNotConnectedException;
import org.java_websocket.handshake.ServerHandshake;

public class SocketClient extends WebSocketClient {

  private final KraSocketClientEventInitiater kraSocketClientEventInitiater;
  @Getter
  private final KraSocketClientProtocol kraProtocol;
  @Getter
  private final List<Long> webSocketAckTimeouts;
  private Timer timer;
  @Getter
  @Setter
  private int dest;

  public SocketClient(URI serverUri, KraSocketClientEventInitiater kraSocketClientEventInitiater,
      KraSocketClientProtocol kraProtocol, List<Long> webSocketAckTimeouts) {
    super(serverUri);

    this.kraSocketClientEventInitiater = kraSocketClientEventInitiater;
    this.kraProtocol = kraProtocol;
    this.webSocketAckTimeouts = webSocketAckTimeouts;
    this.dest = 0;
  }

  public void SendMessage(KraSocketClientProtocolMessage kraProtocolMessage) {
    try {
      if (isOpen()) {
        if (kraProtocolMessage.getStatus() == KraSocketClientProtocolStatus.GET.getStatus()) {
          this.kraProtocol.getSendQueueMessages()
              .put(kraProtocolMessage.getCmdID(), kraProtocolMessage);
        }
        send(this.kraProtocol.EncodeMessage(kraProtocolMessage));
      }
    } catch (WebsocketNotConnectedException exception) {
      KraSocketClient.getLogger().info("SendMessage error");
      exception.printStackTrace();
    }
  }

  public void Reconnect() {
    KraSocketClient.getLogger().info("Reconnecting...");
    reconnect();
  }

  @Override
  public void onOpen(ServerHandshake serverHandshake) {
    this.kraSocketClientEventInitiater.callOnOpen(serverHandshake);
    KraSocketClient.getLogger().info("connection opened");

    TimerTask timerTask = new KraSocketClientMessageTimerTask(this);
    this.timer = new Timer(true);
    this.timer.scheduleAtFixedRate(timerTask, 0, 3 * 1000);
  }

  @Override
  public void onMessage(String s) {

  }

  @Override
  public void onMessage(ByteBuffer byteBuffer) {
    KraSocketClientProtocolMessage kraProtocolMessage = new KraSocketClientProtocol().DecodeMessage(
        byteBuffer.array());

    KraSocketClient.getLogger().info("rec msg: " + kraProtocolMessage.getCmdID() + "cmdNumber"
        + kraProtocolMessage.getCmdNumber());

    if (kraProtocolMessage.getStatus() == KraSocketClientProtocolStatus.ERROR_NO_PERMS.getStatus()
        || kraProtocolMessage.getStatus()
        == KraSocketClientProtocolStatus.ERROR_TRY_AGAIN.getStatus()
        || kraProtocolMessage.getStatus()
        == KraSocketClientProtocolStatus.ERROR_ARG_LEN_TOO_BIG.getStatus()) {

      KraSocketClient.getLogger().warning("error! status: " + kraProtocolMessage.getStatus());
      return;
    }

    // ack
    if (kraProtocolMessage.getStatus() == KraSocketClientProtocolStatus.GET.getStatus()) {
      // response to the user that this message cmdID is already in the queue
      if (this.kraProtocol.getReceivedQueueMessages().containsKey(kraProtocolMessage.getCmdID())) {
        // TODO: Are the status and the cmdID not sufficient here?
        kraProtocolMessage = new KraSocketClientProtocolMessage(
            KraSocketClientProtocolStatus.MESSAGE_ALREADY_IN_QUEUE.getStatus(),
            kraProtocolMessage.getCmdID(), 0,
            UUID.fromString("bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb"),
            (short) kraProtocolMessage.getCmdNumber(), "", null);

        KraSocketClient.getLogger()
            .info("response MESSAGE_ALREADY_IN_QUEUE cmdID: " + kraProtocolMessage.getCmdID()
                + " cmdNumber: " + kraProtocolMessage.getCmdNumber());
        this.SendMessage(kraProtocolMessage);
        return;
      } else { // not in received queue -> add msg to received queue
        this.kraProtocol.getReceivedQueueMessages()
            .put(kraProtocolMessage.getCmdID(), kraProtocolMessage);
      }

      // after the connection is established, the backend sends a Dest, which is assigned here
      if (kraProtocolMessage.getCmdNumber() == 33767) {
        KraSocketClient.getLogger().info("set dest to " + kraProtocolMessage.getArgs());

        this.setDest(Integer.parseInt(kraProtocolMessage.getArgs()));

        this.getKraProtocol().replyThatWeReceivedAckMessage(this, this.getKraProtocol(),
            kraProtocolMessage.getCmdID(),
            KraSocketClientProtocolDest.BACKEND.getStatus(),
            UUID.fromString("bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb"),
            kraProtocolMessage.getCmdNumber());
      }
    }

    this.kraSocketClientEventInitiater.callOnMessage(kraProtocolMessage);
  }

  @Override
  public void onClose(int code, String reason, boolean remote) {
    this.kraSocketClientEventInitiater.callOnClose(code, reason, remote);
    KraSocketClient.logger.warning(
        "connection closed. Code: " + code + " reason: " + reason + " remote: " + remote);

    this.timer.cancel();
  }

  @Override
  public void onError(Exception e) {
    KraSocketClient.getLogger().warning("Err: " + e);
  }
}
